require_relative "operators"
require_relative "rugal"
require_relative "solution"

module GA
    class Algorithm
        def initialize(factory, options={})
            @factory = factory
            
            @selection = options[:selection] || GA::Operators::Selection::RouletteWheel.new
            @crossover = options[:crossover] || GA::Operators::Crossover::SinglePoint.new
            @mutation  = options[:mutation]  || GA::Operators::Mutation::Uniform.new
            @operators = options[:operators] || []
            
            @population_size    = options[:population_size] || 100
            @elite              = options[:elite]           || 0.1
            
            GA.contract("Factory must not be nil") { factory }
            GA.contract("Factory must be a SolutionFactory") { factory.is_a?(GA::Solutions::SolutionFactory) }
            GA.contract("Elite must be non-negative (#@elite given)") { @elite >= 0 }
            GA.contract("All the additional operators must be Operator") { @operators.all? { |op| op.is_a? GA::Operators::Operator } }
        end
    end
    
    class ClassicGenetic < Algorithm
        def initialize(factory, options={})
            super
        end
        
        def maximize(iterations, options = {}, &fitness)
            optimize(iterations, options, &fitness)
        end
        
        def minimize(iterations, options = {}, &fitness)
            optimize(iterations, options) { |chromosome| -fitness.call(chromosome) }
        end
        
        def optimize(iterations, options = {}, &fitness)
            GA.contract("You must pass the fitness function as a block") { fitness }
            optimum = options[:optimum] || nil
            
            GA.contract("Iterations must be more than 0") { iterations & iterations > 0 }
            GA.contract("Optimum must be numeric (or nil)") { !optimum || optimum.is_a?(Numeric) }
            
            population = @factory.population(@population_size)
            
            elite = []
            iterations.times do |i|
                GA::Logger.info "Generation #{i}"
                population.each { |chromosome| chromosome.fitness = fitness.call(chromosome) }
                
                population = population - population.min(elite.size) { |chromosome| chromosome.fitness } + elite 
                
                elite = population.max(@elite * @population_size) { |chromosome| chromosome.fitness }
                if optimum
                    return elite[0] if elite[0].fitness == optimum
                end

                population = @selection.apply(population)
                population = @crossover.apply(population)                
                population =  @mutation.apply(population)
                
                # Uses additional operators
                @operators.each do |operator|
                    population = operator.apply(population)
                end
                
                GA::Logger.info { "Average fitness: #{population.map { |c| c.fitness }.sum.to_f / population.size}" }
            end
            
            return population.max { |chromosome| fitness.call(chromosome) }
        end
    end
    
    class RandomSearch
        def initialize(factory, options = {})
            @factory = factory
            @population_size = options[:population_size] || 1
        end
        
        def maximize(iterations, options = {}, &fitness)
            optimize(iterations, options, &fitness)
        end
        
        def minimize(iterations, options = {}, &fitness)
            optimize(iterations, options) { |chromosome| -fitness.call(chromosome) }
        end
        
        def optimize(iterations, options = {}, &fitness)
            optimum = options[:optimum] || nil
            
            GA.contract("Iterations must be more than 0") { iterations > 0 }
            GA.contract("Optimum must be numeric (or nil)") { !optimum || optimum.is_a?(Numeric) }
            
            best = nil
            iterations.times do |i|
                GA::Logger.info "Generation #{i}"
                population = @factory.population(@population_size)
                population.each { |chromosome| chromosome.fitness = fitness.call(chromosome) }
                
                current_best = population.max { |chromosome| chromosome.fitness}
                best = current_best if !best || current_best.fitness > best.fitness
                
                GA::Logger.info { "Average fitness: #{population.map { |c| c.fitness }.sum.to_f / population.size}" }
            end
            
            return best
        end
    end
end
