require_relative "genetic_algorithm"
require_relative "rugal"
require_relative "solution"

module GA::Operators
    class Operator
    end
end

module GA::Operators::Selection
    class Basic < GA::Operators::Operator
    end
    
    class RouletteWheel < Basic
        def apply(population)
            min_fitness = population.map(&:fitness).min.abs + 1
            fitnesses = population.map { |chromosome| chromosome.fitness != 0 ? chromosome.fitness + min_fitness : 0}
            
            total_fitness = fitnesses.sum
            
            return population.dup if total_fitness == 0
            
            current_total = 0.0
            selection_array = []
            fitnesses.each do |fitness|
                current_total += fitness.to_f / total_fitness
                selection_array.push current_total
            end
            
            offspring = []
            population.size.times do
                value = GA::Random.rand
                selected_index = selection_array.index { |v| value <= v }
                offspring.push population[selected_index].dup
            end
            
            return offspring
        end
    end
    
    class Random < Basic
        def apply(population)
            offspring = []
            population.size.times do
                index = GA::Random.rand(population.size)
                offspring.push population[index].dup
            end
            return offspring
        end
    end
end

module GA::Operators::Crossover
    class Basic < GA::Operators::Operator
        def initialize(options = {})
            @forbidden = []
        end
        
        protected
        def crossover(population)
            result = []
            
            shuffled = GA::Random.shuffle(population)
            shuffled.each_slice(2) do |c1, c2|
                crossed = yield(c1, c2)
                result += crossed
            end
            
            return result
        end
        
        def forbid(type)
            @forbidden.push type
        end
        
        def apply(population)
            check_population(population)
        end
        
        def check_population(population)
            GA.contract("#{self.class} crossover does not accept #{@forbidden.join(", ")}") { population.none? { |c| @forbidden.include?(c.class) } }
        end
    end
    
    class SinglePoint < Basic
        def initialize(options = {})
            super
            
            @probability = options[:probability] || 0.8
            GA.contract("Probability must be in the range [0, 1]") { @probability.between?(0, 1) }
            
            forbid GA::Solutions::PermutationSolution
        end
        
        def apply(population)
            super
            
            crossover(population) do |c1, c2|
                c1 = c1.dup
                c2 = c2.dup
                
                unless c2
                    [c1]
                else
                    if GA::Random.with_probability(@probability)
                        crossover_position = GA::Random.rand([c1.size, c2.size].min)
                        
                        for i in 0...crossover_position
                            c1.data[i], c2.data[i] = c2.data[i], c1.data[i]
                        end
                    end
                    
                    [c1, c2]
                end
            end
        end
    end
    
    class PMX < Basic
        def initialize(options = {})
            super
            
            @probability = options[:probability] || 0.8
            @percent     = options[:size]        || 0.5
            
            GA.contract("Probability must be in the range [0, 1]") { @probability.between?(0, 1) }
            GA.contract("Swath size percentage must be in the range (0, 1)") { @percent > 0 && @percent < 1 }
            
            forbid GA::Solutions::IntegerSolution
            forbid GA::Solutions::FloatSolution
            forbid GA::Solutions::StringSolution
        end
        
        def apply(population)
            super
            
            crossover(population) do |c1, c2|                
                unless c2
                    [c1]
                else
                    t1 = c1
                    t2 = c2
                    if GA::Random.with_probability(@probability)
                        size = GA::Random.rand_int(@percent * c1.size)
                        from = GA::Random.rand_int(c1.size - size)
                        to = from + size
                        
                        c1, c2 = pmx(t1, t2, from, to)
                    end
                    
                    [c1, c2]
                end
            end
        end
        
        private
        def pmx(c1, c2, from, to)
            range = from...to
            
            child1 = c1.dup
            child2 = c2.dup
            
            swath1 = child1.data[range]
            swath2 = child2.data[range]
            
            uswath1 = swath1 - swath2
            uswath2 = swath2 - swath1
            
            mapping1to2 = {}
            mapping2to1 = {}
            for i in 0...uswath1.size
                mapping1to2[uswath1[i]] = uswath2[i]
                mapping2to1[uswath2[i]] = uswath1[i]
            end
                        
            child1.data[range], child2.data[range] = child2.data[range], child1.data[range]
            
            changed_indices = ((0...child1.size).to_a - range.to_a)
            for i in changed_indices
                child1.data[i] = mapping2to1[child1.data[i]] if swath2.include?(child1.data[i])
                child2.data[i] = mapping1to2[child2.data[i]] if swath1.include?(child2.data[i])
            end
            
#             for i in ((0...child1.size).to_a - range.to_a)
#                 v = child2.data[i]
#                 
#                 if swath1.include?(v)
#                     rep = swath2[swath1.index(v)]
#                     child2.data[i] = rep
#                 end
#             end
            
            GA.postcondition("All data in the permutation must be used: #{child1.to_s}") { child1.valid? }
            GA.postcondition("All data in the permutation must be used: #{child2.to_s}") { child2.valid? }
            
            return child1, child2
        end
    end
end

module GA::Operators::Mutation
    class Basic < GA::Operators::Operator
        protected
        def mutation(population)
            result = []
            
            population.each do |chromosome|
                result.push yield(chromosome)
            end
            
            return result
        end
    end
    
    class Uniform < Basic
        def initialize(options = {})
            @probability = options[:probability]  || 0.01
            @radius      = options[:radius]       || 1
            @epsilon     = options[:epsilon]      || 0.01
            
            @epsilon = @epsilon.to_f
            
            GA.contract("Probability must be in the range [0, 1]") { @probability.between?(0, 1) }
            GA.contract("Epsilon must be greater than 0") { @epsilon > 0 }
        end
        
        def apply(population)
            mutation(population) do |chromosome|
                if GA::Random.with_probability(@probability)
                    chromosome = chromosome.mutate(GA::Random.rand(chromosome.size)) do |v| 
                        GA::Random.random_sign * [GA::Random.rand * @radius, @epsilon].max
                    end
                end
                
                chromosome
            end
        end
    end
    
    class Positional < Basic
        def initialize(options = {})
            @probability = options[:probability] || 0.01
            @radius      = options[:radius]      || 1
            @epsilon     = options[:epsilon]     || 1
            
            GA.contract("Probability must be in the range [0, 1]") { @probability.between?(0, 1) }
            GA.contract("Epsilon must be greater than 0") { @epsilon > 0 }
        end
        
        def apply(population)
            mutation(population) do |chromosome|
                if GA::Random.with_probability(@probability)
                    chromosome = chromosome.dup
                    
                    i1 = GA::Random.rand(chromosome.size)
                    i2 = (i1 + [GA::Random.rand(@epsilon+1), @epsilon].max) % chromosome.size
                    
                    chromosome.data[i1], chromosome.data[i2] = chromosome.data[i2], chromosome.data[i1]
                end
                
                chromosome
            end
        end
    end
end

module GA::Operators::Additional
    class Injector < GA::Operators::Operator
        def initialize(factory, options = {})
            @factory = factory
            
            @probability = options[:probability] || 0.01
            @size        = options[:size]        || 1
            
            GA.contract("Factory cannot be nil") { @factory }
            GA.contract("Factory must be a SolutionFactory") { @factory.is_a?(GA::Solutions::SolutionFactory) }
            GA.contract("Probability must be in the range [0, 1]") { @probability.between?(0, 1) }
            GA.contract("Size must be greater than 0") { @size > 0 }
        end
        
        def apply(population)
            if GA::Random.with_probability @probability
                population = population.dup
                size = [GA::Random.rand(@size + 1), 1].max
                
                to_remove = population.sample size
                
                population -= to_remove
                population += @factory.population size
            end
            
            return population
        end
    end
end
