require_relative "operators"
require_relative "rugal"
require_relative "genetic_algorithm"

module GA::Solutions
    class Solution
        attr_accessor   :fitness
        attr_reader     :data
        
        def initialize
            @data = nil
            @fitness = nil
        end
        
        def size
            @data.size
        end
        
        def mutate(i, &block)
            duplicate = self.dup
            duplicate.mutate!(i, &block)
            return duplicate
        end
        
        def mutate!(i)
            @data[i] = neighbour(@data[i], yield(@data[i]))
        end
        
        def neighbour(value, by)
            raise "Cannot infer the neighbours. Must be implemented in #{self.class}."
        end
        
        def dup
            duplicate = super
            duplicate.internal_dup!
            
            return duplicate
        end
        
        def inspect
            "<" + self.class.to_s + " " + @data.inspect + ": #@fitness>"
        end
        
        def to_s
            @data.to_s
        end
        
        protected
        def internal_dup!
            @data = @data.dup
            @fitness
        end
    end
        
    class IntegerSolution < Solution
        def initialize(size, options = {})
            @min = options[:min] || 0
            @max = options[:max] || 999999999
            
            GA.contract("Min must be an Integer") { @min.is_a?(Integer) }
            GA.contract("Max must be an Integer") { @max.is_a?(Integer) }
            GA.contract("Max must be greater than min") { @max > @min }
            
            
            @data = size.times.map { GA::Random.rand(@max+1) + @min }
        end
        
        def neighbour(value, by)
            GA.contract("Value value must be an Integer") { value.is_a?(Integer) }
            GA.contract("By must be a Numeric (#{by.class} given)") { by.is_a?(Numeric) }
            GA.contract("By cannot be 0") { by != 0 }
            by = by > 0 ? by.ceil : by.floor
            
            return [[value + by, @max].min, @min].max
        end
    end
    
    class FloatSolution < Solution
        def initialize(size, options = {})
            @min = (options[:min] || 0.0).to_f
            @max = (options[:max] || 1.0).to_f
            
            GA.contract("Max and min must not be equal") { @max > @min }
            
            @data = size.times.map { ((@max-@min) * GA::Random.rand) + @min }
        end
        
        def neighbour(value, by)
            GA.contract("Value value must be a Float") { value.is_a?(Float) }
            GA.contract("By must be a Numeric (#{by.class} given)") { by.is_a?(Numeric) }
            GA.contract("By cannot be 0") { by != 0 }
            
            return [[value + by, @max].min, @min].max
        end
    end
        
    class StringSolution < Solution
        def initialize(size, options = {})
            @allowed = options[:allowed] || [('A'..'Z'), ('a'..'z'), ('0'..'9')].map { |e| e.to_a }.flatten
            
            GA.contract("Allowed must be an array (#{@allowed.class} given)") { @allowed.is_a?(Array) }
            GA.contract("Allowed must be an array of characters, i.e., 1-sized Strings") { @allowed.all? { |c| c.is_a?(String) && c.size == 1 } }
            GA.contract("There must be at least an allowed character") { @allowed.size > 0 }
            
            @data = size.times.map { @allowed[GA::Random.rand(@allowed.size)]}
        end
        
        def neighbour(value, by)
            GA.contract("Value must be a character (1-sized String)") { value.is_a?(String) && value.length == 1 }
            GA.contract("Value must be an allowed character: <#{@allowed.join(", ")}>") { @allowed.include?(value) }
            GA.contract("By must be a Numeric (#{by.class} given)") { by.is_a?(Numeric) }
            GA.contract("By cannot be 0") { by != 0 }
            
            by = by > 0 ? by.ceil : by.floor
            
            return @allowed[@allowed.index(value) % @allowed.size]
        end
        
        def to_s
            @data.join ""
        end
    end
    
    class PermutationSolution < Solution
        def self.force(data)
            chromosome = self.new(data)
            for i in 0...chromosome.size
                chromosome.data[i] = data[i]
            end
            
            GA.postcondition("Invalid forced chromosome") { chromosome.valid? }
            
            return chromosome
        end
        
        def initialize(content)
            @content = content
            
            GA.contract("Content must not be nil") { @content }
            GA.contract("Content must contain at least an element") { @content.size > 0 }
            
            @data = GA::Random.shuffle @content
        end
        
        def mutate!(index)
            by = yield(@data[index])
            
            switch_with = (index + by) % @data.size
            switch_with = (index + 1) % @data.size if switch_with == index
            
            GA.contract("By must be a Numeric (#{by.class} given)") { by.is_a?(Numeric) }
            GA.contract("By cannot be 0") { by != 0 }
            
            @data[index], @data[switch_with] = @data[switch_with], @data[index]
        end
        
        def valid?
            ((@data - @content) + (@content - @data)) == []
        end
    end
    
    class SolutionFactory
        def initialize(&block)
            @generator = block
        end
        
        def chromosome
            return @generator.call
        end
        
        def population(size)
            size.times.map { chromosome }
        end
    end
end

class Array
    def sample(n)
        GA::Random.shuffle(self)[0...n]
    end
end
